static char rcsid[] = "@(#)$Id: zbuffer.c,v 2.5 1994/11/16 09:19:39 peter Exp $";
/*
 * zbuffer.c - routines for the z-buffer(s).

 * Copyright 1994, 16th November.
 * By Peter Chang
 * peterc@v2.ph.man.ac.uk

 * See notice in misc.h for details of permissions and warranty of this
 * software.

 *
 * $Log: zbuffer.c,v $
 * Revision 2.5  1994/11/16  09:19:39  peter
 * Putting xpgs-2.5 in trunk.
 *
 * Revision 2.0.1.1  1994/11/16  09:10:30  peter
 * Import of xpgs-2.5: archive of new release to alt.sources (11/94)
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define SIRDS_STUFF
#include "pgs.h"
#include "polyh.h"
#include "zbuffer.h"
#include "lookup.h"

#if __GNUC__ == 2
USE(rcsid);
#endif


/* Defines all objects using discretised height fields
 * contains all relatives scales that depend on window height */
int sfillall(int yy, int **zbufa, int **zbufb, float mf, Affine *gt)
{
   int i;
   int x, y;
   int *zll, *zlr;
   float z;
   int bv, iv;

   void sfillobj(Object *, float , float , float , int **zta, int **ztb,
		 Affine *gt);

   iv = (reverse > 0) ? MYHIGH : MYLOW;

   buf_top = yy;
   buf_bot = yy + buf_h;
   for (y=0; y<buf_h; y++) { /* set background to const initially */
      zll = zbufa[y];
      zlr = zbufb[y];
      for (x=0; x<width; x++) {
         *zll++ = iv;
         *zlr++ = iv;
      }
   }

   for (i=0; i<nobj; i++) sfillobj(&shapes[i], mf, mf, mf, zbufa, zbufb, gt);

   z = OS - reverse*background*OY;
   bv = (int) ((enlarge*(1 << fineness))*ES*z/(VD+reverse*z));

   for (y=0; y<buf_h; y++) { /* reset background */
      zll = zbufa[y];
      zlr = zbufb[y];
      for (x=0; x<width; x++, zll++, zlr++) {
         if (*zll == iv) *zll = bv;
         if (*zlr == iv) *zlr = bv;
      }
   }
   return bv;
}


int zfillall(int yy, int **zbuf, float mf, Affine *gt)
{
   int i;
   int x, y;
   int *zl;
   int bv, iv;

   void zfillobj(Object *, float , float , float , int **zt, Affine *gt);

   iv = MYLOW;

   buf_top = yy;
   buf_bot = yy + buf_h;
   for (y=0; y<buf_h; y++) { /* set background to const initially */
      zl = zbuf[y];
      for (x=0; x<width; x++) {
         *zl++ = iv;
      }
   }

   for (i=0; i<nobj; i++) zfillobj(&shapes[i], mf, mf, mf, zbuf, gt);

   bv = (int) ((enlarge*(1 << fineness))*OY*background);

   for (y=0; y<buf_h; y++) { /* reset background */
      zl = zbuf[y];
      for (x=0; x<width; x++, zl++) {
         if (*zl == iv) *zl = bv;
      }
   }
   return bv;
}


/*
 * Dumps height data to two 's'-buffers
 */
int zsfillall(int yy, int *zmap, int **zbufa, int **zbufb, int ttw, int tth,
	      int xs, int ys)
{
   int x, y;
   int *zll, *zlr;
   float z;
   int bv, iv;
   int half;

   void zsdrawobj(int *, int **, int **, int , int , int , int , int );

   iv = (reverse > 0) ? MYHIGH : MYLOW;
   buf_top = yy;
   buf_bot = yy + buf_h;
   for (y=0; y<buf_h; y++) { /* set background to const initially */
      zll = zbufa[y];
      zlr = zbufb[y];
      for (x=0; x<width; x++) {
         *zll++ = iv;
         *zlr++ = iv;
      }
   }

   half = (buf_bot < height/2) ? 0 : 1;
   if (half == 1 && yy < height/2) half = 2;
#ifdef MYDEBUG
   printf("yy %d buf t %d %d %d : half %d\n", yy,
	  buf_top, buf_bot, buf_h, half);
#endif
   zsdrawobj(zmap, zbufa, zbufb, ttw, tth, xs, ys, half);

   z = OS - reverse*background*OY;
   bv = (int) ((enlarge*(1 << fineness))*ES*z/(VD+reverse*z));
   for (y=0; y<buf_h; y++) { /* reset background */
      zll = zbufa[y];
      zlr = zbufb[y];
      for (x=0; x<width; x++, zll++, zlr++) {
         if (*zll == iv) *zll = bv;
         if (*zlr == iv) *zlr = bv;
      }
   }
   return bv;
}


/*
 * a triangle, coords relative to centre of window and in pixels
 * new improved algorithm

 * take the canonical triangle

 *                     --3
 *                ----' '
 *            2--'   --'
 *           /    --'
 *          /  --'
 *         /--'
 *         1

 *  and split it into two

 *            2------                  --3
 *           /    --'             ----' /  B
 *          /  --'  A         2--'-----/
 *         /--'
 *         1

 *   then scan from point 1 upwards line by line to fill A
 *   and do triangle B in the opposite direction from point 3.
 */

/*
 * comp = 1 -> reversed buffer
 * comp = 0 -> normal buffer
 */

#if defined(__DECC) || defined(__osf)
#pragma inline (triangle)
#endif

static __inline__ void triangle (int x1, int y1, float z1,
				 int x2, int y2, float z2,
               int x3, int y3, float z3, int **zbuffer, int comp)
{
   int   i, end, swapped;
   int   x12, x13, y12, y13, y23;
   float ztemp, temp;
   int   xtemp, ytemp;
   float xbeg, xend;
   float dxb, dxe;
   float dzdx, dzdy, zsh;
   int *zline;

/* buffer height, top and bottom (in screen coords) */

   /* sort points in increasing y */
   if (y1 > y2) {
      xtemp=x1; ytemp=y1; ztemp=z1;
      x1=x2; y1=y2; z1=z2;
      x2=xtemp; y2=ytemp; z2=ztemp;
   }
   if (y1 > y3) {
      xtemp=x1; ytemp=y1; ztemp=z1;
      x1=x3; y1=y3; z1=z3;
      x3=xtemp; y3=ytemp; z3=ztemp;
   }
   if (y2 > y3) {
      xtemp=x3; ytemp=y3; ztemp=z3;
      x3=x2; y3=y2; z3=z2;
      x2=xtemp; y2=ytemp; z2=ztemp;
   }

   x12 = x2-x1;
   x13 = x3-x1;
   y12 = y2-y1;
   y13 = y3-y1;
   y23 = y3-y2;
   xend = z2-z1;
   xbeg = z3-z1;

   /* work out the partial derivatives in x and y  of z = f(x,y) */
   zsh = (float) (x12 * y13 - y12 * x13);
   if (zsh == 0.0) return;       /* points project on to a line */
   zsh = 1.0/zsh;
   dzdx = (xend * y13 - y12 * xbeg)*zsh;
   dzdy = (x12 * xbeg - xend * x13)*zsh;

   /*
    * find the gradients of the two edges of A and make sure the
    * beginning of the line is on the left
    */
   swapped = 0;
   dxe = x13 / (float) y13;
   if (y12 == 0) {
      if (x12 > 0) {
         swapped = 1;
         dxb = dxe;
      } else
         dxb = 1e6;
   } else {
      dxb = x12 / (float) y12;
      if (dxb > dxe) {
         xend = dxb;
         dxb = dxe;
         dxe = xend;
         swapped = 1;
      }
   }

   /* initialise variables */
   x1 += OX;
   zsh = dxb * dzdx + dzdy;
   xbeg = (float) x1;
   xend = (float) x1;

   /* clip triangle A in vertical direction, if necessary */
   ytemp = OY - y1;
   if (ytemp < buf_top) return;      /* entire thing out of view */
   if (ytemp >= buf_bot) {
      ytemp -= buf_bot - 1;
      xbeg += dxb * ytemp;
      xend += dxe * ytemp;
      z1 += zsh * ytemp;
      y12 -= ytemp;
      ytemp = buf_bot - 1;
   }
   if (ytemp - y12 < buf_top) y12 = ytemp - buf_top + 1; /* A too tall */

   /* double loop to fill A, with various checks for lateral clips */
   for (i=0; i<y12; i++) {
      end = (int) (xend  + 1.5);
      if (end > width) end = width;
      if (xbeg < 0) {
         ztemp = z1 - dzdx*xbeg;
         xtemp = 0;
      } else {
         ztemp = z1;
         xtemp = (int) xbeg;
      }
      if (xbeg >= width || xend < 0) end = 0;
      zline = &zbuffer[ytemp%buf_h][xtemp];
      for (; xtemp<end; xtemp++) {
	 temp = ztemp - *zline;
	 if ((temp > 0.0) ^ comp) *zline = (int) (ztemp+0.5);
         ztemp += dzdx;
         zline++;
      }
      xbeg += dxb;
      xend += dxe;
      z1 += zsh;
      ytemp--;
   }


   /*
    * find the gradient of final edge of B and make sure the
    * beginning of the line is on the left
    */
   if (y23 == 0)
      ztemp = 1e6;
   else
      ztemp = (x3 - x2) / (float) y23;
   if (swapped)
      dxe = ztemp;
   else
      dxb = ztemp;

   /* initialise variables */
   x3 += OX;
   zsh = dxb * dzdx + dzdy;
   xbeg = (float) x3;
   xend = (float) x3;

   /* clip triangle B */
   ytemp = OY - y3;
   if (ytemp < buf_top) {
      /*
       * in a previous attempt, I got this wrong.
       * much thanks to Luke Hutchison (lukeh@kcbbs.gen.nz) for
       * pointing this out.
       * the error was due to us going down (instead of up) the screen
       * and so the gradients should be multiplied by a minus sign
       * but in the section ytemp is also negative.
       */
      ytemp -= buf_top;
      xbeg += dxb*ytemp;
      xend += dxe*ytemp;
      z3 += zsh*ytemp;
      y23 += ytemp;
      ytemp = buf_top;
   }
   if (ytemp >= buf_bot) return; /* B out of view */
   if (y23 + ytemp >= buf_bot) y23 = buf_bot - ytemp - 1; /* B too tall */

   /* double loop to fill B */
   y23++;
   for (i=0; i<y23; i++) {
      end = (int) (xend + 1.5);
      if (end > width) end = width;
      if (xbeg < 0) {
         ztemp = z3 - dzdx*xbeg;
         xtemp = 0;
      } else {
         ztemp = z3;
         xtemp = (int) xbeg;
      }
      if (xbeg >= width || xend < 0) end = 0;
      zline = &zbuffer[ytemp%buf_h][xtemp];
      for (; xtemp<end; xtemp++) {
	 temp = ztemp - *zline;
	 if ((temp > 0.0) ^ comp) *zline = (int) (ztemp+0.5);
         ztemp += dzdx;
         zline++;
      }
      xbeg -= dxb;
      xend -= dxe;
      z3 -= zsh;
      ytemp++;
   }
}


/*
 * an object,
 * made from its polygonal faces
 * hidden surfaces are removed using a z-buffer
 * polygons are constructed from triangles
 */

void sfillobj(Object *pobj, float xm, float ym, float zm,
	      int **zbufa, int **zbufb, Affine *gt)
{
   int i, j;
   int xa, ya, xb, yb, xc, yc;
   float za, zb, zc, zt;
   float efn, es, oy, os, vd, vdos;
   int cflag;
   usint *polygon;
   float *lkup;
   Affine taf;
   Point tpt;

   es = ES;
   es *= reverse;
   os = OS;
   lkup = zplookup;
   lkup += (int) ((reverse>0) ? ZMUL*height : ZMUL*os);
   os *= reverse;
   vd = VD;
   vdos = vd + os;
   efn = enlarge*(1 << fineness);

   /*
    * points are given in normalised coordinates,
    * we scale it all up relative to the centre of the window and
    * window height
    */

   oy = height/2.0;
   xm *= oy;
   ym *= oy;
   zm *= oy;
   taf = pobj->wt;
   mulaffine(&taf, gt);

   cflag = (1+reverse)/2; /* = 1 for wide, = 0 for cross-eyed */
   i=0;
   polygon = pobj->polygon;
   do {
      j = i + 1 + *polygon++;
      applyaffine(&tpt, &pobj->ri[*polygon++], &taf);
      zt = zm*tpt.z;
      switch(cflag) {
       case 0:
	 if (zt < os) zt = os; /* os is actually -os */
	 else if (zt >= vdos) zt = vdos - 0.01*oy;
	 break;
       case 1: default:
	 if (zt > os) zt = os;
	 else if (zt < -height) zt = (float) -height;
	 break;
      }
/*      zt = vd/(vdos - zt); */
      zt = *(lkup+(int)(ZMUL*zt));
      xa = (int) (zt*xm*tpt.x);
      ya = (int) (zt*ym*tpt.y);
      za = es*(1 - zt);
      xa -= za/2.0;
      i += 3;
      applyaffine(&tpt, &pobj->ri[*polygon++], &taf);
      zt = zm*tpt.z;
      switch(cflag) {
       case 0:
	 if (zt < os) zt = os;
	 else if (zt >= vdos) zt = vdos - 0.01*oy;
	 break;
       case 1: default:
	 if (zt > os) zt = os;
	 else if (zt < -height) zt = (float) -height;
	 break;
      }
      zt = *(lkup+(int)(ZMUL*zt));
      xb = (int) (zt*xm*tpt.x);
      yb = (int) (zt*ym*tpt.y);
      zb = es*(1 - zt);
      xb -= zb/2.0;
      do {
	 i++;
	 applyaffine(&tpt, &pobj->ri[*polygon++], &taf);
	 zt = zm*tpt.z;
	 switch(cflag) {
	    case 0:
	    if (zt < os) zt = os;
	    else if (zt >= vdos) zt = vdos - 0.01*oy;
	    break;
	    case 1: default:
	    if (zt > os) zt = os;
	    else if (zt < -height) zt = (float) -height;
	    break;
	 }
	 zt = *(lkup+(int)(ZMUL*zt));
	 xc = (int) (zt*xm*tpt.x);
	 yc = (int) (zt*ym*tpt.y);
	 zc = es*(1 - zt);
	 xc -= zc/2.0;
         triangle(xa, ya, efn*za, xb, yb, efn*zb,
		  xc, yc, efn*zc, zbufa, cflag);
         triangle((int) (xa+za), ya, efn*za, (int) (xb+zb), yb, efn*zb,
		  (int) (xc+zc), yc, efn*zc, zbufb, cflag);
	 xb = xc;
	 yb = yc;
	 zb = zc;
      } while (i<j);
   } while (i<pobj->nindex);
}


void zfillobj(Object *pobj, float xm, float ym, float zm, int **zbuf,
	      Affine *gt)
{
   int i, j;
   int xa, ya, xb, yb, xc, yc;
   float za, zb, zc, zt;
   float efn, oy, os;
   usint *polygon;
   Affine taf;
   Point tpt;

   os = OS;
   efn = enlarge*(1 << fineness);

   oy = height/2.0;
   xm *= oy;
   ym *= oy;
   zm *= oy;
   taf = pobj->wt;
   mulaffine(&taf, gt);

   i=0;
   polygon = pobj->polygon;
   do {
      j = i + 1 + *polygon++;
      applyaffine(&tpt, &pobj->ri[*polygon++], &taf);
      zt = zm*tpt.z;
      if (zt > os) zt = os;
      else if (zt < -oy) zt = -oy;
      xa = (int) (xm*tpt.x);
      ya = (int) (ym*tpt.y);
      za = efn*zt;
      i += 3;
      applyaffine(&tpt, &pobj->ri[*polygon++], &taf);
      zt = zm*tpt.z;
      if (zt > os) zt = os;
      else if (zt < -oy) zt = -oy;
      xb = (int) (xm*tpt.x);
      yb = (int) (ym*tpt.y);
      zb = efn*zt;
      do {
	 i++;
	 applyaffine(&tpt, &pobj->ri[*polygon++], &taf);
	 zt = zm*tpt.z;
	 if (zt > os) zt = os;
	 else if (zt < -oy) zt = -oy;
	 xc = (int) (xm*tpt.x);
	 yc = (int) (ym*tpt.y);
	 zc = efn*zt;
         triangle(xa, ya, za, xb, yb, zb, xc, yc, zc, zbuf, 0);
	 xb = xc;
	 yb = yc;
	 zb = zc;
      } while (i<j);
   } while (i<pobj->nindex);
}


/*
 * scan z-buffer and use a MxN square of z-values to
 * fill 's'-buffers
 */
void zsdrawobj(int *zmap, int **zbufl, int **zbufr,
	       int ttw, int tth, int xstep, int ystep, int half)
{
   int xa, ya, xb, yb, xc, yc, xd, yd;
   float sa, sb, sc, sd;
   float efn, es, ox, oy, os;
   int cflag;
   float z[4];
   int *zuline, *zdline;
   int l, m, n, p;
   float zm, zl, zt;
   int x, y, yt, rx, ry;
   float *lkup;

#define AZERO 0.01

   es = ES;
   es *= reverse;
/*   vd = VD; */
   os = OS;
/*   vdos = vd + os; */
   efn = enlarge*(1 << fineness);
   oy = tth/2.0;
   ox = ttw/2.0;
   lkup = zplookup;
   lkup += (int) ((reverse>0) ? ZMUL*height : ZMUL*os);

   cflag = (1+reverse)/2;
   switch(half) {
    case 0:
      ry = oy;
      y = ystep;
      yt = tth/2;
      break;
    case 1:
      ry = -ystep;
      y = tth/2;
      yt = tth;
      break;
    case 2: default:
      ry = oy;
      y = ystep;
      yt = tth;
      break;
   }
   for (; y<yt; ry-=ystep, y+=ystep) {
      rx = -ox;
      zuline = &zmap[(y-ystep)*ttw];
      zdline = &zmap[y*ttw];
      z[0] = *zuline;
      zuline += xstep;
/*      zt = vd/(vdos - z[0]); */
      zt = *(lkup+(int)(ZMUL*z[0]));
      xa = zt*rx;
      ya = zt*ry;
      sa = es*(1 - zt);
      xa -= sa/2;
      z[3] = *zdline;
      zdline += xstep;
      zt = *(lkup+(int)(ZMUL*z[3]));
      xd = zt*rx;
      yd = zt*(ry-ystep);
      sd = es*(1 - zt);
      xd -= sd/2;
      for (x=xstep; x<ttw; x+=xstep) {
	 rx += xstep;
	 z[1] = *zuline;
	 zuline += xstep;
	 zt = *(lkup+(int)(ZMUL*z[1]));
	 xb = zt*rx;
	 yb = zt*ry;
	 sb = es*(1 - zt);
	 xb -= sb/2;
	 z[2] = *zdline;
	 zdline += xstep;
	 zt = *(lkup+(int)(ZMUL*z[2]));
	 xc = zt*rx;
	 yc = zt*(ry-ystep);
	 sc = es*(1 - zt);
	 xc -= sc/2;

	 if (fabs(z[0]-z[1]+z[2]-z[3]) < AZERO) {
            /* planar case, 4 vertices */
	    m = 1;
	 } else {
	    /* non-planar case -> two triangle */
            /* find most and least */
            l = m = 0;
            zl = zm = z[0];
            for (n=1; n<4; n++) {
               zt = z[n];
               if (zt > zm) {
                  zm = zt;
                  m = n;
               } else if (zt < zl) {
                  zl = zt;
                  l = n;
               }
	    }
            /* check if there are common heights */
            switch(l-m) {
             case -3: case -1: /* min clockwise adjacent from max */
               p = (m+1) & 3;
               if ((zm-z[p]) < AZERO) m = l;
               break;
             case -2: case 2:  /* min opposite to max */
               break;
             case 1: case 3:   /* min clockwise adjacent from max */
               p = (l+1) & 3;
               if ((z[p]-zl) > AZERO) m = l;
               break;
             case 0: default:
               break;
            }
	 }

	 if ((m&1) == 1) { /* start at corner a */
	    triangle(xa, ya, efn*sa, xb, yb, efn*sb,
		     xc, yc, efn*sc, zbufl, cflag);
	    triangle(xa, ya, efn*sa, xc, yc, efn*sc,
		     xd, yd, efn*sd, zbufl, cflag);
	    triangle((xa+sa), ya, efn*sa, (xb+sb), yb, efn*sb,
		     (xc+sc), yc, efn*sc, zbufr, cflag);
	    triangle((xa+sa), ya, efn*sa, (xc+sc), yc, efn*sc,
		     (xd+sd), yd, efn*sd, zbufr, cflag);
         } else {  /* end at corner a */
	    triangle(xd, yd, efn*sd, xa, ya, efn*sa,
		     xb, yb, efn*sb, zbufl, cflag);
	    triangle(xd, yd, efn*sd, xb, yb, efn*sb,
		     xc, yc, efn*sc, zbufl, cflag);
	    triangle((xd+sd), yd, efn*sd, (xa+sa), ya, efn*sa,
		     (xb+sb), yb, efn*sb, zbufr, cflag);
	    triangle((xd+sd), yd, efn*sd, (xb+sb), yb, efn*sb,
		     (xc+sc), yc, efn*sc, zbufr, cflag);
         }

	 z[0] = z[1];
	 xa = xb;
	 ya = yb;
	 sa = sb;
	 z[3] = z[2];
	 xd = xc;
	 yd = yc;
	 sd = sc;
      }
   }
}


/* memory allocation for zbuffer */
int **zmema(void)
{
   int i;
   int **zbuf;

   DIEIF((zbuf = (int **) malloc(buf_h*sizeof(int *))) == NULL,
         "Unable to allocate memory for z-buffer\n");
   for (i=0; i<buf_h; i++) {
      DIEIF((zbuf[i] = (int *) malloc(width*sizeof(int))) == NULL,
            "Unable to allocate memory for z-buffer\n");
   }
   return zbuf;
}


void zmemf(int **zbuf)
{
   int i;

   if (zbuf != NULL) {
      for(i=0; i<buf_h; i++) free(zbuf[i]);
      free(zbuf);
   }
}
